#include <a8/a8.h>
#include <a8/awaiter.h>
#include <a8/promise.h>

#include <f8/f8.h>
#include <f8/coroutine.h>
#include <f8/comgr.h>
#include <f8/timer.h>

#ifdef USE_BOOST

class TimerPromise : public a8::Promise
{
 public:

    TimerPromise(long long time) { time_ = time; };

protected:

    virtual void DoAwait() override
    {
        f8::Timer::Instance()->SetTimeout
            (
             time_,
             [this, _self = shared_from_this()] (int event, const a8::Args* args) mutable
             {
                 if (event == a8::TIMER_EXEC_EVENT) {
                     DoDone();
                 }
             });
    }

 private:
    long long time_ = 0;

};

namespace f8
{

    Coroutine::Coroutine(std::function<void(Coroutine*)> cb)
    {
        INIT_LIST_HEAD(&co_entry_);
        INIT_LIST_HEAD(&exec_entry_);
        list_add_tail(&co_entry_, &f8::CoMgr::Instance()->co_list_);
        Attach();
        cb_ = cb;
        source_ = std::make_shared<boost::coroutines2::coroutine<void>::pull_type>
            (
             [this] (boost::coroutines2::coroutine<void>::push_type& sink)
             {
                 CallEnter(sink);
                 cb_(this);
                 DoDone();
                 CallExit(sink);
             });
    }

    Coroutine::~Coroutine()
    {
        Deatch();
        list_del_init(&co_entry_);
    }

    bool Coroutine::Exec()
    {
        if (*source_) {
            (*source_)();
            return true;
        } else {
            return false;
        }
    }

    void Coroutine::CoSuspend()
    {
        Deatch();
    }

    void Coroutine::CoResume()
    {
        Attach();
    }

    void Coroutine::CoYield()
    {
        (*sink_)();
    }

    std::shared_ptr<a8::Results> Coroutine::CoAwait(std::shared_ptr<a8::Awaiter> awaiter)
    {
        CoSuspend();
        awaiter->Await(shared_from_this());
        while (!awaiter->Done()) {
            CoYield();
        }
        return awaiter->GetResult();
    }

    void Coroutine::Attach()
    {
        list_add_tail(&exec_entry_, &CoMgr::Instance()->exec_list_);
    }

    void Coroutine::Deatch()
    {
        if (!list_empty(&exec_entry_)) {
            list_del_init(&exec_entry_);
        }
    }

    void Coroutine::CallEnter(boost::coroutines2::coroutine<void>::push_type& sink)
    {
        sink_ = &sink;
        CoYield();
    }

    void Coroutine::CallExit(boost::coroutines2::coroutine<void>::push_type& sink)
    {
        sink_ = nullptr;
    }

    void Coroutine::DoResume()
    {
        CoResume();
    }

    std::shared_ptr<a8::Awaiter> Coroutine::Sleep(long long time)
    {
        return std::make_shared<TimerPromise>(time);
    }

}

#endif
