#include <assert.h>

#include <mutex>

#include <a8/a8.h>

#include <f8/f8.h>

#include <f8/timer.h>

#include <f8/msgqueue.h>

namespace f8
{
    struct MsgQueueNode
    {
        struct list_head entry;
        MsgHandleFunc func;
    };

    struct IMMsgNode
    {
        int msgid;
        const a8::Args args;
        IMMsgNode* next = nullptr;

        IMMsgNode(const a8::Args& args1):args(std::move(args1))
        {
        }

    };

    class MsgQueueImp
    {
    public:
        std::map<int, list_head> msg_handlers;

        std::mutex im_msg_mutex_;
        IMMsgNode* im_top_node_ = nullptr;
        IMMsgNode* im_bot_node_ = nullptr;
        IMMsgNode* im_work_node_ = nullptr;

        ~MsgQueueImp()
        {
            for (auto& pair : msg_handlers) {
                while (!list_empty(&pair.second)) {
                    MsgQueue::Instance()->RemoveCallBack(pair.second.next);
                }
            }
            im_msg_mutex_.lock();
            if (!im_work_node_) {
                im_work_node_ = im_top_node_;
                im_top_node_ = nullptr;
                im_bot_node_ = nullptr;
            }
            while (im_work_node_) {
                IMMsgNode* pdelnode = im_work_node_;
                im_work_node_ = im_work_node_->next;
                delete pdelnode;
                if (!im_work_node_) {
                    im_work_node_ = im_top_node_;
                    im_top_node_ = nullptr;
                    im_bot_node_ = nullptr;
                }
            }
            im_msg_mutex_.unlock();
        }

        void Update()
        {
            if (!im_work_node_ && im_top_node_) {
                im_msg_mutex_.lock();
                im_work_node_ = im_top_node_;
                im_top_node_ = nullptr;
                im_bot_node_ = nullptr;
                im_msg_mutex_.unlock();
            }
            while (im_work_node_) {
                IMMsgNode *pdelnode = im_work_node_;
                ProcessMsg(pdelnode->msgid, pdelnode->args);
                im_work_node_ = im_work_node_->next;
                delete pdelnode;
            }
        }

        void ProcessMsg(int msgid, const a8::Args& args)
        {
            auto itr = msg_handlers.find(msgid);
            if (itr != msg_handlers.end()) {
                list_head* head = &itr->second;
                struct MsgQueueNode *node = nullptr;
                struct MsgQueueNode *tmp = nullptr;
                list_for_each_entry_safe(node, tmp, head, entry) {
                    node->func(args);
                }
            }
        }

        CallBackHandle RegisterCallBack(int msgid, MsgHandleFunc handle_func)
        {
            MsgQueueNode* node = new MsgQueueNode();
            INIT_LIST_HEAD(&node->entry);
            node->func = handle_func;

            auto itr = msg_handlers.find(msgid);
            if (itr == msg_handlers.end()) {
                msg_handlers[msgid] = list_head();
                itr = msg_handlers.find(msgid);
                assert(itr != msg_handlers.end());
                INIT_LIST_HEAD(&itr->second);
            }
            list_add_tail(&node->entry, &itr->second);
            return &node->entry;
        }

        void PostMsg(int msgid, const a8::Args args)
        {
            IMMsgNode *p = new IMMsgNode(args);
            p->msgid = msgid;
            p->next = nullptr;
            im_msg_mutex_.lock();
            if (im_bot_node_) {
                im_bot_node_->next = p;
                im_bot_node_ = p;
            } else {
                im_top_node_ = p;
                im_bot_node_ = p;
            }
            im_msg_mutex_.unlock();
#if 0
            NotifyLoopCond();
#endif
        }

        bool HasMsg()
        {
            if (!im_work_node_) {
                im_msg_mutex_.lock();
                if (!im_work_node_ && im_top_node_) {
                    im_work_node_ = im_top_node_;
                    im_top_node_ = nullptr;
                    im_bot_node_ = nullptr;
                }
                im_msg_mutex_.unlock();
            }
            if (im_work_node_) {
                return true;
            }
            return false;
        }

    };

    void MsgQueue::Init()
    {
        imp_ = std::make_shared<MsgQueueImp>();
    }

    void MsgQueue::UnInit()
    {
    }

    void MsgQueue::Update()
    {
        imp_->Update();
    }

    bool MsgQueue::HasMsg()
    {
        return imp_->HasMsg();
    }

    void MsgQueue::RemoveCallBack(CallBackHandle handle)
    {
        list_head* head = handle;
        MsgQueueNode* node = list_entry(head, struct MsgQueueNode, entry);
        list_del_init(&node->entry);
        delete node;
    }

    CallBackHandle MsgQueue::RegisterCallBack(int msgid, MsgHandleFunc handle_func)
    {
        return imp_->RegisterCallBack(msgid, handle_func);
    }

    void MsgQueue::PostMsg(int msgid, const a8::Args args)
    {
        imp_->PostMsg(msgid, std::move(args));
    }

}
